/*
 GBxCart RW - Wonderswan Backup Save
 Version: 1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 7/11/2016
 Last Modified: 27/01/2023
 License: CC-BY-NC-SA
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <signal.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

void sigintHandler(int sig_num) {
	set_mode('0');
	RS232_flushRX(cport_nr);
	gbx_cart_power_down();
	exit(0);
}

int main(int argc, char **argv) {
	signal(SIGINT, sigintHandler);
	
	printf("GBxCart RW - WS Adapter Backup Save v1.1 by insideGadgets\n");
	printf("#########################################################\n");
	
	read_config();
	
	// Open COM port
	if (com_test_port() == 0) {
		printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
		read_one_letter();
		return 1;
	}
	printf("Connected on COM port: %i at ", cport_nr+1);
	
	// Break out of any existing functions on ATmega
	set_mode('0');
	RS232_flushRX(cport_nr);
	
	// Get cartridge mode - Gameboy or Gameboy Advance
	cartridgeMode = request_value(CART_MODE);
	
	// Get PCB version
	gbxcartPcbVersion = request_value(READ_PCB_VERSION);
	xmas_wake_up();
	
	// Increase to 1.7Mbit baud for v1.4 PCB
	#ifdef _WIN32
	if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == MINI_1_1) {
		set_mode(USART_1_7M_SPEED);
		RS232_CloseComport(cport_nr);
		delay_ms(200);
		
		if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
			printf("1.7M Baud\n");
		}
	}
	else {
		printf("1M Baud\n");
	}
	#else
		printf("1M Baud\n");
	#endif
	
	char pcbVersionString[20];
	if (gbxcartPcbVersion == PCB_1_0) {
		strncpy(pcbVersionString, "v1.0", 5);
	}
	else if (gbxcartPcbVersion == PCB_1_1) {
		strncpy(pcbVersionString, "v1.1/v1.2", 10);
	}
	else if (gbxcartPcbVersion == PCB_1_3) {
		strncpy(pcbVersionString, "v1.3", 5);
	}
	else if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS)) {
		strncpy(pcbVersionString, "v1.4", 5);
	}
	else if (gbxcartPcbVersion == MINI) {
		strncpy(pcbVersionString, "Mini", 5);
	}
	else if (gbxcartPcbVersion == MINI_1_1) {
		strncpy(pcbVersionString, "Mini v1.1", 10);
	}
	else if (gbxcartPcbVersion == GBXMAS) {
		strncpy(pcbVersionString, "XMAS", 5);
	}
	
	// Get firmware version
	gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
	printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
	
	set_mode(VOLTAGE_3_3V);
	
	// GBx v1.4 - Power up the cart if not already powered up and flush buffer
	gbx_cart_power_up();
	RS232_flushRX(cport_nr);
	
	// Get cartridge mode - Gameboy or Gameboy Advance
	cartridgeMode = request_value(CART_MODE);
	
	// Wonderswan setup
	set_mode(WS_SETUP);
	delay_ms(200);
	
	
	// Get the date/time to be used for the filename
	time_t rawtime;
	struct tm* timeinfo;
	char timebuffer[80];
	
	time (&rawtime);
	timeinfo = localtime (&rawtime);
	strftime (timebuffer,80,"Wonderswan_%Y-%m-%d_%H-%M-%S.sav",timeinfo);
	
	// Create a new file
	FILE *romFile = fopen(timebuffer, "wb");
	
	
	printf("\nBacking up Wonderswan save to %s\n", timebuffer);
	printf("[             25%%             50%%             75%%            100%%]\n[");
	
	uint32_t currentAddress = 0x10000;
	uint32_t endAddress = 0x18000;
	uint32_t readBytes = 0;
	set_number(currentAddress, SET_START_ADDRESS);
	set_mode(WS_READ_RAM);
	
	while (currentAddress < endAddress) {	
		uint8_t comReadBytes = com_read_bytes(romFile, 64);
		if (comReadBytes == 64) {
			currentAddress += 64;
			readBytes += 64;
			if (currentAddress < endAddress) {
				com_read_cont();
			}
			print_progress_percent(readBytes, 32768 / 64);
			led_progress_percent(readBytes, 32768 / 28);
		}
	}
	com_read_stop();
	fclose(romFile);
	
	printf("\n\nCompleted");
	
	read_one_letter();
	
	return 0;
}